// parser.cpp

#include "parser.h"
#include "xmlparser.h"
#include <glib.h>
#include <curl/curl.h>
#include <string.h>
#include <stdlib.h>

// date-time

static const char *months[] = {
	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

std::string format_date (int year, int month, int day, int hour, int min, int hour_zone, int min_zone)
{
	gchar *str = g_strdup_printf ("%s %d, %d (%02d:%02d %c%02d:%02d)",
		months[month-1], day, year, hour, min, hour_zone < 0 ? '-' : '+', abs (hour_zone), min_zone);
	std::string s (str);
	g_free (str);
	return s;
}

std::string parse_rfc3339 (const std::string &text)  // Atom
{
	// spec: http://tools.ietf.org/html/rfc3339#section-5.6
	if (text.size() < 20)
		return "";
	const char *str = text.c_str();
	int year, month, day, hour, min, hour_zone, min_zone;
	if (sscanf (str, "%4d-%2d-%2dT%2d:%2d", &year, &month, &day, &hour, &min) == 5) {
		str = str + 19;
		if (*str == '.')  // time-secfrac
			do { ++str;
			} while (*str != 'Z' && *str != '+' && *str != '-' && *str != '\0');
		if (*str == '\0')
			return "";
		if (*str == 'Z')
			hour_zone = min_zone = 0;
		else {
			sscanf (str+1, "%2d:%2d", &hour_zone, &min_zone);
			if (*str == '-')
				hour_zone = -hour_zone;
		}
		return format_date (year, month, day, hour, min, hour_zone, min_zone);
	}
	return "";
}

std::string parse_rfc822 (const std::string &text)  // RSS
{
	// spec: http://asg.web.cmu.edu/rfc/rfc822.html
	if (text.size() < 26)
		return "";
	int year, month, day, hour, min, hour_zone, min_zone;
	const char *str = text.c_str();
	str += 5;
	day = atoi (str);
	str += 3;
	for (month = 0; month < 12; month++)
		if (!strncmp (str, months[month], 3))
			break;
	if (month == 12)
		return "";
	month += 1;
	str += 4;
	year = atoi (str);
	str += 5;
	hour = atoi (str);
	str += 3;
	min = atoi (str);
	str += 6;
	if (!strncmp (str, "GMT", 3))
		hour_zone = min_zone = 0;
	else {
		sscanf (str+1, "%2d%2d", &hour_zone, &min_zone);
		if (*str == '-')
			hour_zone = -hour_zone;
	}
	return format_date (year, month, day, hour, min, hour_zone, min_zone);
}

// download

static char curl_errorBuffer [CURL_ERROR_SIZE] = { 0 };

bool download (const std::string &url, write_callback func, void *data)
{
	curl_errorBuffer[0] = '\0';
	CURL *curl;
	CURLcode result;
	curl = curl_easy_init();
	if (curl) {
		std::string buffer;
		curl_easy_setopt(curl, CURLOPT_ERRORBUFFER, curl_errorBuffer);
		curl_easy_setopt(curl, CURLOPT_URL, url.c_str());
		curl_easy_setopt(curl, CURLOPT_HEADER, 0);
		curl_easy_setopt(curl, CURLOPT_FOLLOWLOCATION, 1);
		curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, func);
		curl_easy_setopt(curl, CURLOPT_WRITEDATA, data);
		result = curl_easy_perform (curl);  // action
		curl_easy_cleanup(curl);  
		return result == CURLE_OK;
	}
	return false;
}

struct DownloadData
{
	char *url;
	write_callback func;
	void *data;

	DownloadData (const char *url, write_callback func, void *data)
	: url (g_strdup (url)), func (func), data (data) {}
	~DownloadData()
	{ g_free (url); }
};

static gpointer _download_thread (gpointer _data)
{
	DownloadData *data = (DownloadData *) _data;
	download (data->url, data->func, data->data);
	delete data;
	return 0;
}

void download_thread (const char *url, write_callback func, void *data)
{
	DownloadData *pdata = new DownloadData (url, func, data);
	GThread *thread = g_thread_create_full (_download_thread, pdata, 0, FALSE, FALSE,
		G_THREAD_PRIORITY_LOW, NULL);
	if (!thread)
		download (url, func, data);
}

std::string download (const std::string &url)
{
	struct inner {
		static size_t writer (char *data, size_t size, size_t nitems, void *user_data)
		{
			std::string *buffer = (std::string *) user_data;
			size_t len = size * nitems;
			buffer->append (data, len);
			return len;
		}
	};

	std::string buffer;
	if (download (url, inner::writer, &buffer))
		return buffer;
	return "";
}

//** RSS

struct RssItemParser : public XmlParser::Handler  // <item>
{
	RssItemParser (ParseNewsHandler *handler)
	: handler (handler) {}

private:
	ParseNewsHandler *handler;

	virtual XmlParser::Handler *startElement (const char *name,
		const char **attribute_names, const char **attribute_values,
		std::string &error) { return NULL; }

	virtual void textElement (const char *name, const std::string &text, std::string &error)
	{
		if (!strcmp (name, "title"))
			handler->setTitle (text);
		else if (!strcmp (name, "link"))
			handler->setLink (text);
		else if (!strcmp (name, "content") || !strcmp (name, "content:encoded"))
			handler->setContent (text);
		else if (!strcmp (name, "description"))
			handler->setSummary (text);
		else if (!strcmp (name, "pubDate")) {
			std::string date (parse_rfc822 (text));
			handler->setDate (date);
		}
		else if (!strcmp (name, "author"))
			handler->setAuthor (text);
		else if (!strcmp (name, "category"))
			handler->addCategory (text);
		else if (!strcmp (name, "guid"))
			handler->setId (text);
	}

	virtual void endElement (const char *name, XmlParser::Handler *child, std::string &error)
	{ delete child; }
};

struct RssChannelParser : public XmlParser::Handler  // <channel>
{
	RssChannelParser (ParseFeedHandler *handler)
	: handler (handler) {}

private:
	ParseFeedHandler *handler;

	virtual XmlParser::Handler *startElement (const char *name,
		const char **attribute_names, const char **attribute_values,
		std::string &error)
	{
		if (!strcmp (name, "item")) {
			ParseNewsHandler *newsHandler = handler->appendNews();
			return new RssItemParser (newsHandler);
		}
		return NULL;
	}

	virtual void textElement (const char *name, const std::string &text, std::string &error)
	{
		if (!strcmp (name, "title"))
			handler->setTitle (text);
		else if (!strcmp (name, "link"))
			handler->setLink (text);
		else if (!strcmp (name, "description"))
			handler->setDescription (text);
		else if (!strcmp (name, "managingEditor"))
			handler->setAuthor (text);
	}

	virtual void endElement (const char *name, XmlParser::Handler *child, std::string &error)
	{ delete child; }
};

struct RssParser : public XmlParser::Handler  // <rss>
{
	RssParser (ParseFeedHandler *handler)
	: handler (handler) {}

private:
	ParseFeedHandler *handler;

	virtual XmlParser::Handler *startElement (const char *name,
		const char **attribute_names, const char **attribute_values,
		std::string &error)
	{
		if (!strcmp (name, "channel"))
			return new RssChannelParser (handler);
		return NULL;
	}

	virtual void textElement (const char *name, const std::string &text, std::string &error)
	{
		if (!strcmp (name, "title"))
			handler->setTitle (text);
		else if (!strcmp (name, "link"))
			handler->setLink (text);
		else if (!strcmp (name, "description"))
			handler->setDescription (text);
	}

	virtual void endElement (const char *name, XmlParser::Handler *child, std::string &error)
	{ delete child; }
};

//** Atom

struct AtomEntryAuthorParser : public XmlParser::Handler  // <author>
{
	AtomEntryAuthorParser (ParseNewsHandler *handler)
	: handler (handler) {}

private:
	ParseNewsHandler *handler;

	virtual XmlParser::Handler *startElement (const char *name,
		const char **attribute_names, const char **attribute_values,
		std::string &error)
	{ return NULL; }

	virtual void textElement (const char *name, const std::string &text, std::string &error)
	{
		if (!strcmp (name, "name"))
			handler->setAuthor (text);
	}

	virtual void endElement (const char *name, XmlParser::Handler *child, std::string &error) {}
};

struct AtomEntryParser : public XmlParser::Handler  // <entry>
{
	AtomEntryParser (ParseNewsHandler *handler)
	: handler (handler) {}

private:
	ParseNewsHandler *handler;

	virtual XmlParser::Handler *startElement (const char *name,
		const char **attribute_names, const char **attribute_values,
		std::string &error)
	{
		if (!strcmp (name, "link")) {
			const char *type = XmlParser::get_value ("type",
				attribute_names, attribute_values);
			const char *href = XmlParser::get_value ("href",
				attribute_names, attribute_values);
			if (href && (!type || !strcmp (type, "text/html")))
				handler->setLink (href);
		}
		else if (!strcmp (name, "author"))
			return new AtomEntryAuthorParser (handler);
		return NULL;
	}

	virtual void textElement (const char *name, const std::string &text, std::string &error)
	{
		if (!strcmp (name, "title"))
			handler->setTitle (text);
		else if (!strcmp (name, "summary"))
			handler->setSummary (text);
		else if (!strcmp (name, "content"))
			handler->setContent (text);
		else if (!strcmp (name, "created") || !strcmp (name, "published")) {
			std::string date (parse_rfc3339 (text));
			handler->setDate (date);
		}
		else if (!strcmp (name, "category") || !strcmp (name, "dc:subject"))
			handler->addCategory (text);
		else if (!strcmp (name, "id"))
			handler->setId (text);
	}

	virtual void endElement (const char *name, XmlParser::Handler *child, std::string &error)
	{ delete child; }
};

struct AtomAuthorParser : public XmlParser::Handler  // <author>
{
	AtomAuthorParser (ParseFeedHandler *handler)
	: handler (handler) {}

private:
	ParseFeedHandler *handler;

	virtual XmlParser::Handler *startElement (const char *name,
		const char **attribute_names, const char **attribute_values,
		std::string &error)
	{ return NULL; }

	virtual void textElement (const char *name, const std::string &text, std::string &error)
	{
		if (!strcmp (name, "name"))
			handler->setAuthor (text);
	}

	virtual void endElement (const char *name, XmlParser::Handler *child, std::string &error) {}
};

struct AtomFeedParser : public XmlParser::Handler  // <feed>
{
	AtomFeedParser (ParseFeedHandler *handler)
	: handler (handler) {}

private:
	ParseFeedHandler *handler;

	virtual XmlParser::Handler *startElement (const char *name,
		const char **attribute_names, const char **attribute_values,
		std::string &error)
	{
		if (!strcmp (name, "entry"))
			return new AtomEntryParser (handler->appendNews());
		else if (!strcmp (name, "author"))
			return new AtomAuthorParser (handler);
		else if (!strcmp (name, "link")) {
			const char *type = XmlParser::get_value ("type",
				attribute_names, attribute_values);
			const char *href = XmlParser::get_value ("href",
				attribute_names, attribute_values);
			if (href && (!type || !strcmp (type, "text/html")))
				handler->setLink (href);
		}
		return NULL;
	}

	virtual void textElement (const char *name, const std::string &text, std::string &error)
	{
		if (!strcmp (name, "title"))
			handler->setTitle (text);
		else if (!strcmp (name, "subtitle"))
			handler->setDescription (text);
	}

	virtual void endElement (const char *name, XmlParser::Handler *child, std::string &error)
	{ delete child; }
};

//** top XML block

struct TopParser : public XmlParser::Handler  // <rss>
{
	TopParser (ParseFeedHandler *handler)
	: handler (handler) {}

private:
	ParseFeedHandler *handler;

	virtual XmlParser::Handler *startElement (const char *name,
		const char **attribute_names, const char **attribute_values,
		std::string &error)
	{
		if (!strcmp (name, "rss")) {
			const char *version = XmlParser::get_value ("version",
				attribute_names, attribute_values);
			if (!strcmp (version, "1.0"))
				error = std::string ("Unsupported RSS version: ") + version;
			else
				return new RssParser (handler);
		}
		else if (!strcmp (name, "feed"))
			return new AtomFeedParser (handler);
		else
			error = std::string ("Unsupported format: ") + name;
		return NULL;
	}

	virtual void textElement (const char *name, const std::string &text,
		std::string &error) {}
	virtual void endElement (const char *name, XmlParser::Handler *child, std::string &error)
	{ delete child; }
};

struct ParseThreadData {
	ParseFeedHandler *handler;
	std::string url, codeset;
	parse_done_callback callback;

	ParseThreadData (ParseFeedHandler *handler, const std::string &url,
	                 const std::string &codeset, parse_done_callback callback)
	: handler (handler), url (url), codeset (codeset), callback (callback)
	{}
};

static gpointer parse_thread (gpointer _data)
{
	ParseThreadData *data = (ParseThreadData *) _data;
	GError *error = 0;

	std::string text = download (data->url);
	if (text.empty()) {
		const char *msg = "Download failed";
		if (curl_errorBuffer[0])
			msg = &(curl_errorBuffer[0]);
		data->callback (data->handler, msg);
		return 0;
	}
	if (!data->codeset.empty()) {
		gsize bytes_read, bytes_written;
		gchar *_text = g_convert (text.c_str(), -1, "utf8", data->codeset.c_str(),
		                          &bytes_read, &bytes_written, &error);
		if (!_text) {
			data->callback (data->handler, error->message);
			g_error_free (error);
			return 0;
		}
		text = _text;
		g_free (_text);
	}

	std::string error_msg;
	{
		TopParser handler (data->handler);
		XmlParser parser (&handler);
		parser.parse (text, error_msg);
	}

	data->callback (data->handler, error_msg);
	delete data;
	return 0;
}

void parse (ParseFeedHandler *handler, const std::string &url,
            const std::string &codeset, parse_done_callback callback)
{
	ParseThreadData *data = new ParseThreadData (handler, url, codeset, callback);

	GError *error;
	GThread *thread = g_thread_create_full (parse_thread, data, 0, FALSE, FALSE,
		G_THREAD_PRIORITY_LOW, &error);
	if (!thread)
		printf ("Couldn't create thread for parser: %s\n", error->message);
}

